import os, sys

def byte_line(data):
	line = "    db "
	for i in range(15):
		line += "${:02X}, ".format(data[i+0])
	return line + "${:02X}\n".format(data[15+0])

def word_line(data):
	line = "    dw "
	for i in range(15):
		line += "${:04X}, ".format(data[i*2+1]*256 + data[i*2+0])
	return line + "${:04X}\n".format(data[15*2+1]*256 + data[15*2+0])

def create_map_bank_file(bank, bank_num, path):
	file_name = "bank_" + "{:03x}".format(bank_num) + ".asm"
	
	file_content = "; Bank " + "{:X}".format(bank_num) + " Level Data\n"
	file_content += ";  Extracted using extract.py\n"
	file_content += "SECTION \"ROM Bank $" + "{:03X}".format(bank_num) + "\", ROMX[$4000], BANK[$" + "{:X}".format(bank_num) + "]"
	file_content += "\n; Screen Data Pointers\n\n"
	for i in range(0x0000, 0x0200, 0x20):
		file_content += word_line(bank[i:i+32])
	file_content += "\n; Scroll Data\n\n"
	for i in range(0x0200, 0x0300, 0x10):
		file_content += byte_line(bank[i:i+16])
	file_content += "\n; Room Transition Indexes\n\n"
	for i in range(0x0300, 0x0500, 0x20):
		file_content += word_line(bank[i:i+32])
	for s in range(0x0500, 0x4000, 0x0100):
		file_content += "\n; Screen ${:04X}\n\n".format(0x4000+s)
		for i in range(s, s+0x0100, 0x10):
			file_content += byte_line(bank[i:i+16])
	
	with open(os.path.join(path, file_name), "w") as f:
		f.write(file_content)
	
def main():
    if len(sys.argv) < 2 or len(sys.argv) > 4:
        sys.exit("usage: " + sys.argv[0] + " <romfile> [<bank_start=9> [<bank_end=F>]]")
    file_name = sys.argv[1]
    path = os.path.dirname(file_name)
    bank_start = int(sys.argv[2], 16) if len(sys.argv) >= 3 else 0x9
    bank_end = int(sys.argv[3], 16) if len(sys.argv) >= 4 else 0xF
    
    bank_list = []
    with open(file_name, "rb") as f:
	    for bank_num in range(bank_end+1):
		    bank_list.append(f.read(0x4000));

    for bank_num in range(bank_start, bank_end+1):
	    create_map_bank_file(bank_list[bank_num], bank_num, path)
    print("Success!")
    
if __name__ == "__main__":
    main()
