import sys, os, traceback
from PySide6 import QtCore, QtWidgets, QtGui
from PIL import Image, ImageQt, UnidentifiedImageError
from interface.EnemyEdit import Interface_EnemyEdit

"""
KEYWORDS:
blockset = set of possible blocks and/or the interface to edit this
map = block placements and camera restraints and transitions in a bank
world = bank data + enemies and/or the interface to edit this

"""

class Interface_MapEdit(QtWidgets.QMainWindow):
    def __init__(self):
        super().__init__()

        # variables
        self.src_folder = "../../SRC/"
        
        self.blockset_data = None
        self.blockset_file_name = None

        self.map_data_arrangement = None
        self.map_data_scroll = None
        self.map_data_transition = None
        self.map_data_screen = None
        self.map_data_bank_number = None
        self.map_file_name = None

        self.world_x = 7
        self.world_y = 7
        self.selected_block = 0
        
        self.enemy_names = None
        self.enemy_ids = None
        self.enemy_banks = None

        # controls
        self.menubar = self.menuBar()

        self.menubar_file = QtWidgets.QMenu()
        self.menubar_file.setTitle("File")
        self.menubar.addMenu(self.menubar_file)

        self.menubar_file_load_blockset = QtGui.QAction("Load blockset image...")
        self.menubar_file.addAction(self.menubar_file_load_blockset)

        self.menubar_file_load_map_file = QtGui.QAction("Load map file...")
        self.menubar_file.addAction(self.menubar_file_load_map_file)

        self.menubar_file_save_map_file = QtGui.QAction("Save map file...")
        self.menubar_file.addAction(self.menubar_file_save_map_file)
        
        self.menubar_file_save_enemy_file = QtGui.QAction("Save enemy file...")
        self.menubar_file.addAction(self.menubar_file_save_enemy_file)
        
        self.menubar_edit = QtWidgets.QMenu()
        self.menubar_edit.setTitle("Edit")
        self.menubar.addMenu(self.menubar_edit)

        self.menubar_edit_actiongroup = QtGui.QActionGroup(self)

        self.menubar_edit_map = QtGui.QAction("Map", checkable=True, checked=True)
        self.menubar_edit_actiongroup.addAction(self.menubar_edit_map)
        self.menubar_edit.addAction(self.menubar_edit_map)
        
        self.menubar_edit_arrangement = QtGui.QAction("Screen Arrangement", checkable=True)
        self.menubar_edit_actiongroup.addAction(self.menubar_edit_arrangement)
        self.menubar_edit.addAction(self.menubar_edit_arrangement)
        
        self.menubar_edit_scroll = QtGui.QAction("Scroll", checkable=True)
        self.menubar_edit_actiongroup.addAction(self.menubar_edit_scroll)
        self.menubar_edit.addAction(self.menubar_edit_scroll)

        self.menubar_edit_enemy = QtGui.QAction("Enemy", checkable=True)
        self.menubar_edit_actiongroup.addAction(self.menubar_edit_enemy)
        self.menubar_edit.addAction(self.menubar_edit_enemy)
        
        self.menubar_view = QtWidgets.QMenu()
        self.menubar_view.setTitle("View")
        self.menubar.addMenu(self.menubar_view)

        self.menubar_view_enemy = QtGui.QAction("Enemy", checkable=True, checked=True)
        self.menubar_view.addAction(self.menubar_view_enemy)
        
        self.root = QtWidgets.QGridLayout()
        self.root.setColumnMinimumWidth(0, 20)
        self.root.setColumnMinimumWidth(1, 512)
        self.root.setColumnMinimumWidth(2, 20)
        self.root.setRowStretch(0, 1)
        self.root.setRowMinimumHeight(1, 256)
        self.root.setRowMinimumHeight(2, 10)
        self.root.setRowStretch(3, 1)
        self.root.setRowMinimumHeight(4, 20)
        self.root.setRowMinimumHeight(5, 512)
        self.root.setRowMinimumHeight(6, 20)
        self.root_widget = QtWidgets.QWidget()
        self.root_widget.setLayout(self.root)
        self.setCentralWidget(self.root_widget)

        self.root_blockset_title = QtWidgets.QLabel()
        self.root_blockset_title.setWordWrap(True)
        self.root.addWidget(self.root_blockset_title, 0, 1)

        self.root_blockset_view = QtWidgets.QGraphicsView()
        self.root_blockset_view.setFixedSize(512, 256)
        self.root_blockset_view.setViewportMargins(-2, -2, -2, -2)
        self.root.addWidget(self.root_blockset_view, 1, 1, alignment=QtCore.Qt.AlignmentFlag.AlignCenter)

        self.root_blockset_scene = QtWidgets.QGraphicsScene()
        self.root_blockset_view.setScene(self.root_blockset_scene)

        self.root_world_title = QtWidgets.QLabel()
        self.root_world_title.setWordWrap(True)
        self.root.addWidget(self.root_world_title, 3, 1)

        self.root_world_btn_move_up = QtWidgets.QPushButton("^")
        self.root_world_btn_move_up.setSizePolicy(QtWidgets.QSizePolicy.Policy.Ignored, QtWidgets.QSizePolicy.Policy.Ignored)
        self.root.addWidget(self.root_world_btn_move_up, 4, 1)

        self.root_world_btn_move_left = QtWidgets.QPushButton("<")
        self.root_world_btn_move_left.setSizePolicy(QtWidgets.QSizePolicy.Policy.Ignored, QtWidgets.QSizePolicy.Policy.Ignored)
        self.root.addWidget(self.root_world_btn_move_left, 5, 0)

        self.root_world_view = QtWidgets.QGraphicsView()
        self.root_world_view.setFixedSize(512, 512)
        self.root_world_view.setViewportMargins(-2, -2, -2, -2)
        self.root_world_view.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self.root_world_view.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self.root.addWidget(self.root_world_view, 5, 1, alignment=QtCore.Qt.AlignmentFlag.AlignCenter)

        self.root_world_scene = QtWidgets.QGraphicsScene()
        self.root_world_view.setScene(self.root_world_scene)

        self.root_world_btn_move_right = QtWidgets.QPushButton(">")
        self.root_world_btn_move_right.setSizePolicy(QtWidgets.QSizePolicy.Policy.Ignored, QtWidgets.QSizePolicy.Policy.Ignored)
        self.root.addWidget(self.root_world_btn_move_right, 5, 2)

        self.root_world_btn_move_down = QtWidgets.QPushButton("v")
        self.root_world_btn_move_down.setSizePolicy(QtWidgets.QSizePolicy.Policy.Ignored, QtWidgets.QSizePolicy.Policy.Ignored)
        self.root.addWidget(self.root_world_btn_move_down, 6, 1)
	
        # init function calls
        if not os.path.exists(self.src_folder):
            self.src_folder = QtWidgets.QFileDialog.getExistingDirectory(caption="Load SRC folder")
            if self.src_folder == "":
                messageBox = QtWidgets.QMessageBox(self)
                messageBox.setWindowTitle("Failed to load SRC folder")
                messageBox.setText("No SRC folder was chosen")
                messageBox.setIcon(QtWidgets.QMessageBox.Icon.Critical)
                messageBox.exec()
                sys.exit()
        
        self.enemy_file_load()
        
        # controller
        self.menubar_file_load_blockset.triggered.connect(self.blockset_file_load)
        self.root_blockset_scene.mousePressEvent = self.blockset_on_click
        self.menubar_file_load_map_file.triggered.connect(self.map_file_load)
        self.menubar_file_save_map_file.triggered.connect(self.map_file_save)
        self.menubar_file_save_enemy_file.triggered.connect(self.enemy_file_save)
        self.menubar_edit_actiongroup.triggered.connect(self.world_update_interface)
        self.menubar_view_enemy.triggered.connect(self.world_update_interface)
        self.root_world_scene.mousePressEvent = self.world_on_click
        self.root_world_scene.mouseMoveEvent = self.world_on_drag
        self.root_world_btn_move_up.clicked.connect(lambda: self.world_position_set(self.world_x, self.world_y - 1))
        self.root_world_btn_move_down.clicked.connect(lambda: self.world_position_set(self.world_x, self.world_y + 1))
        self.root_world_btn_move_left.clicked.connect(lambda: self.world_position_set(self.world_x - 1, self.world_y))
        self.root_world_btn_move_right.clicked.connect(lambda: self.world_position_set(self.world_x + 1, self.world_y))


    def update_interface(self):
        self.blockset_update_interface()
        self.world_update_interface()


    def blockset_file_load(self):
        file_name, _ = QtWidgets.QFileDialog.getOpenFileName(caption="Load blockset", dir=self.src_folder + "/tilesets")
        if file_name == "":
            return
        try:
            blockset_img = Image.open(file_name)
            if blockset_img.size[0] % 16 != 0 or blockset_img.size[1] % 16 != 0:
                raise Exception("image dimensions should be multiples of 16, but are " + str(blockset_img.size))
            
            self.blockset_data = []
            for i in range(0x80):
                x = i % (blockset_img.size[0] / 16)
                y = i // (blockset_img.size[0] / 16)
                if y < blockset_img.size[1] / 16:
                    self.blockset_data.append(blockset_img.copy().crop((x*16, y*16, (x+1)*16, (y+1)*16)))
                else:
                    break
            
            self.blockset_file_name = file_name
            self.update_interface()
        except Exception as ex:
            messageBox = QtWidgets.QMessageBox(self)
            messageBox.setWindowTitle("Failed to load blockset image")
            messageBox.setText(str(ex))
            messageBox.setDetailedText(str(traceback.format_exc()))
            messageBox.setIcon(QtWidgets.QMessageBox.Icon.Critical)
            messageBox.exec()


    def blockset_update_interface(self):
        if self.blockset_file_name is None:
            self.root_blockset_title.setText("Blockset: None loaded...")
            return
        abspath_SRC = os.path.abspath(self.src_folder)
        abspath_blockset = os.path.abspath(self.blockset_file_name)
        abspath_SRC = "/".join(abspath_SRC.split("/")[:-2]) + "/"
        self.root_blockset_title.setText("Blockset: " + abspath_blockset.replace(abspath_SRC, ""))
        
        for i in range(0x80):
            x = i % 16
            y = i // 16
            if i < len(self.blockset_data):
                qimage = ImageQt.ImageQt(self.blockset_data[i])
                pixmap = QtGui.QPixmap.fromImage(qimage).scaled(32, 32)
                pixmap_item = QtWidgets.QGraphicsPixmapItem()
                pixmap_item.setPixmap(pixmap)
                pixmap_item.setPos(x * 32, y * 32)
                self.root_blockset_scene.addItem(pixmap_item)
            else:
                break


    def blockset_on_click(self, mouse_event):
        x = mouse_event.scenePos().x() // 32
        y = mouse_event.scenePos().y() // 32
        self.selected_block = int(x + y * 16)


    def world_update_interface(self):
        if self.map_file_name is None:
            self.root_world_title.setText("Map: None loaded...")
            return
        abspath_SRC = os.path.abspath(self.src_folder)
        abspath_map = os.path.abspath(self.map_file_name)
        abspath_SRC = "/".join(abspath_SRC.split("/")[:-2]) + "/"
        self.root_world_title.setText("Map: " + abspath_map.replace(abspath_SRC, ""))
        
        if self.blockset_file_name is None:
            return
        
        for item in self.root_world_scene.items():
            self.root_world_scene.removeItem(item)

        for y in range(2):
            for x in range(2):
                self.map_draw_screen((x*16, y*16), self.map_data_arrangement[self.world_y + y][self.world_x + x])
        
        if self.menubar_view_enemy.isChecked():
            font = QtGui.QFont('monospace', 6, QtGui.QFont.Bold)
            pen = QtGui.QPen(QtGui.QColor(255, 255, 255))
        
            for y in range(2):
                for x in range(2):
                    enemy_screen = self.enemy_banks[self.map_data_bank_number][self.world_y + y][self.world_x + x]
                    if enemy_screen == None:
                        continue
                    for enemy in enemy_screen:
                        x_pos = enemy["x"] + x*256
                        y_pos = enemy["y"] + y*256
                        self.root_world_scene.addLine(x_pos-10, y_pos, x_pos, y_pos, pen)
                        self.root_world_scene.addLine(x_pos, y_pos-10, x_pos, y_pos+10, pen)
                        self.world_draw_text("${:02X} ".format(enemy["respawn"]), x_pos, y_pos-12, font)
                        self.world_draw_text(enemy["name"], x_pos, y_pos-2, font)
        
        if self.menubar_edit_arrangement.isChecked():
            font = QtGui.QFont('monospace', 12, QtGui.QFont.Normal)
            pen = QtGui.QPen(QtGui.QColor(255, 255, 255))
            brush = QtGui.QBrush(QtGui.QColor(128, 128, 128, 128))
            
            for y in range(2):
                for x in range(2):
                    for xx in range(4):
                        for yy in range(3):
                            self.world_draw_rect(x*256+xx*48+32, y*256+yy*48+32, 48, 48, pen, brush)
                            
                    self.world_draw_text("+$01", x*256+32, y*256+12+32, font)
                    self.world_draw_text("+$02", x*256+80, y*256+12+32, font)
                    self.world_draw_text("+$04", x*256+32, y*256+12+80, font)
                    self.world_draw_text("+$08", x*256+80, y*256+12+80, font)
                    self.world_draw_text("+$10", x*256+32, y*256+12+128, font)
                    self.world_draw_text("+$20", x*256+80, y*256+12+128, font)
                    
                    self.world_draw_text("-$01", x*256+128, y*256+12+32, font)
                    self.world_draw_text("-$02", x*256+176, y*256+12+32, font)
                    self.world_draw_text("-$04", x*256+128, y*256+12+80, font)
                    self.world_draw_text("-$08", x*256+176, y*256+12+80, font)
                    self.world_draw_text("-$10", x*256+128, y*256+12+128, font)
                    self.world_draw_text("-$20", x*256+176, y*256+12+128, font)
                    
                    self.world_draw_text("${:02X}00".format(self.map_data_arrangement[self.world_y + y][self.world_x + x]), x*256+96, y*256+12+176, font)
        
        if self.menubar_edit_scroll.isChecked():
            pen = QtGui.QPen(QtGui.QColor(255, 255, 255))
            brush_lock = QtGui.QBrush(QtGui.QColor(128, 128, 128, 192))
            brush_free = QtGui.QBrush(QtGui.QColor(0, 0, 0, 0))
            
            for y in range(2):
                for x in range(2):
                    scroll_screen = self.map_data_scroll[self.world_y + y][self.world_x + x]
                    
                    # right
                    self.world_draw_rect(x*256+256-32, y*256-16, 16+32, 256+16+16, pen, brush_lock if scroll_screen & 0x1 else brush_free)
                    # left
                    self.world_draw_rect(x*256-16, y*256-16, 16+32, 256+16+16, pen, brush_lock if scroll_screen & 0x2 else brush_free)
                    # up
                    self.world_draw_rect(x*256-16, y*256-16, 256+16+16, 16+32, pen, brush_lock if scroll_screen & 0x4 else brush_free)
                    # down
                    self.world_draw_rect(x*256-16, y*256+256-32, 256+16+16, 16+32, pen, brush_lock if scroll_screen & 0x8 else brush_free)
    
    
    def world_draw_text(self, string, x, y, font, color=QtGui.QColor(255, 255, 255)):
        name_text = self.root_world_scene.addText(string, font)
        name_text.setPos(x, y)
        name_text.setDefaultTextColor(color)
    
    
    def world_draw_rect(self, x, y, w, h, pen, brush):
        x2 = x+w
        y2 = y+h
        x = min(max(x, 1), 256*2-2)
        y = min(max(y, 1), 256*2-2)
        x2 = min(max(x2, 1), 256*2-2)
        y2 = min(max(y2, 1), 256*2-2)
        w = x2-x
        h = y2-y
        rect = QtCore.QRectF(x, y, w, h)
        self.root_world_scene.addRect(rect, pen, brush)
    
    
    def world_on_click(self, mouse_event):
        if self.menubar_edit_map.isChecked():
            self.map_edit(mouse_event)
        elif self.menubar_edit_arrangement.isChecked():
            self.arrangement_edit(mouse_event)
        elif self.menubar_edit_scroll.isChecked():
            self.scroll_edit(mouse_event)
        elif self.menubar_edit_enemy.isChecked():
            self.enemy_edit(mouse_event)


    def world_on_drag(self, mouse_event):
        if mouse_event.buttons() == QtCore.Qt.NoButton:
            return
        if self.menubar_edit_map.isChecked():
            self.map_edit(mouse_event)


    def world_position_set(self, x, y):
        x = min(max(x, 0), 14)
        y = min(max(y, 0), 14)
        self.world_x = x
        self.world_y = y
        self.root_world_btn_move_up.setEnabled(self.world_y > 0)
        self.root_world_btn_move_down.setEnabled(self.world_y < 16 - 2)
        self.root_world_btn_move_left.setEnabled(self.world_x > 0)
        self.root_world_btn_move_right.setEnabled(self.world_x < 16 - 2)
        self.world_update_interface()


    def map_file_load(self):
        file_name, _ = QtWidgets.QFileDialog.getOpenFileName(caption="Load map file", dir=self.src_folder + "/maps", filter="ASM File (*.asm)")
        if file_name == "":
            return
        try:
            with open(file_name, "r") as file:
                map_file = file.readlines()
            bank_number_line = next(line.strip() for line in map_file if "BANK[$" in line)
            self.map_data_bank_number = bank_number_line[bank_number_line.index("BANK[$")+6:-1]
            map_file = [line.strip() for line in map_file]
            map_file = list(filter(lambda x: x.startswith("db") or x.startswith("dw"), map_file))

            self.map_data_arrangement = []
            for i in range(16):
                line = map_file[i+0][3:].split(",")
                line = [word.replace("$", "").strip() for word in line]
                self.map_data_arrangement.append([])
                for word in line:
                    word = int(word[:2], 16)
                    self.map_data_arrangement[i].append(word)

            self.map_data_scroll = []
            for i in range(16):
                line = map_file[i+16][3:].split(",")
                line = [byte.replace("$", "").strip() for byte in line]
                self.map_data_scroll.append([])
                for byte in line:
                    byte = int(byte, 16)
                    self.map_data_scroll[i].append(byte)

            self.map_data_transition = []
            for i in range(16):
                line = map_file[i+32][3:].split(",")
                line = [word.replace("$", "").strip() for word in line]
                self.map_data_transition.append([])
                for word in line:
                    word = int(word, 16)
                    self.map_data_transition[i].append(word)

            self.map_data_screen = {}
            for j in range(0x45, 0x80):
                self.map_data_screen[j] = []
                for i in range(16):
                    line = map_file[i+48+(j-0x45)*16][3:].split(",")
                    line = [byte.replace("$", "").strip() for byte in line]
                    self.map_data_screen[j].append([])
                    for byte in line:
                        byte = int(byte, 16)
                        self.map_data_screen[j][i].append(byte)

            self.map_file_name = file_name
            self.update_interface()
        except Exception as ex:
            messageBox = QtWidgets.QMessageBox(self)
            messageBox.setWindowTitle("Failed to load map file")
            messageBox.setText(str(ex))
            messageBox.setDetailedText(str(traceback.format_exc()))
            messageBox.setIcon(QtWidgets.QMessageBox.Icon.Critical)
            messageBox.exec()


    def map_draw_screen(self, pos, screen_id):
        screen = self.map_data_screen[screen_id]
        for y in range(16):
            for x in range(16):
                block = screen[y][x]
                qimage = ImageQt.ImageQt(self.blockset_data[block])
                pixmap = QtGui.QPixmap.fromImage(qimage).scaled(16, 16)
                pixmap_item = QtWidgets.QGraphicsPixmapItem()
                pixmap_item.setPixmap(pixmap)
                pixmap_item.setPos((x + pos[0]) * 16, (y + pos[1]) * 16)
                self.root_world_scene.addItem(pixmap_item)


    def map_edit(self, mouse_event):
        if self.map_data_screen is None:
            return
        x = int(mouse_event.scenePos().x() // 16)
        y = int(mouse_event.scenePos().y() // 16)
        screen = self.map_data_arrangement[int(self.world_y + y // 16)][int(self.world_x + x // 16)]
        if self.map_data_screen[screen][y % 16][x % 16] != self.selected_block:
            self.map_data_screen[screen][y % 16][x % 16] = self.selected_block
            self.world_update_interface()
    
    
    def arrangement_edit(self, mouse_event):
        for y in range(2):
            for x in range(2):
                mouse_x = int(mouse_event.scenePos().x() // 16) - x*16
                mouse_y = int(mouse_event.scenePos().y() // 16) - y*16
                pos = (self.world_x + x, self.world_y + y)
                
                if mouse_x >= 2 and mouse_x < 5 and mouse_y >= 2 and mouse_y < 5:
                    self.arrangement_add(pos, 0x1)
                if mouse_x >= 5 and mouse_x < 8 and mouse_y >= 2 and mouse_y < 5:
                    self.arrangement_add(pos, 0x2)
                if mouse_x >= 2 and mouse_x < 5 and mouse_y >= 5 and mouse_y < 8:
                    self.arrangement_add(pos, 0x4)
                if mouse_x >= 5 and mouse_x < 8 and mouse_y >= 5 and mouse_y < 8:
                    self.arrangement_add(pos, 0x8)
                if mouse_x >= 2 and mouse_x < 5 and mouse_y >= 8 and mouse_y < 11:
                    self.arrangement_add(pos, 0x10)
                if mouse_x >= 5 and mouse_x < 8 and mouse_y >= 8 and mouse_y < 11:
                    self.arrangement_add(pos, 0x20)
                
                if mouse_x >= 8 and mouse_x < 11 and mouse_y >= 2 and mouse_y < 5:
                    self.arrangement_add(pos, -0x1)
                if mouse_x >= 11 and mouse_x < 14 and mouse_y >= 2 and mouse_y < 5:
                    self.arrangement_add(pos, -0x2)
                if mouse_x >= 8 and mouse_x < 11 and mouse_y >= 5 and mouse_y < 8:
                    self.arrangement_add(pos, -0x4)
                if mouse_x >= 11 and mouse_x < 14 and mouse_y >= 5 and mouse_y < 8:
                    self.arrangement_add(pos, -0x8)
                if mouse_x >= 8 and mouse_x < 11 and mouse_y >= 8 and mouse_y < 11:
                    self.arrangement_add(pos, -0x10)
                if mouse_x >= 11 and mouse_x < 14 and mouse_y >= 8 and mouse_y < 11:
                    self.arrangement_add(pos, -0x20)
        self.world_update_interface()
    
    
    def arrangement_add(self, pos, add):
        new_arrangement = self.map_data_arrangement[pos[1]][pos[0]] + add
        new_arrangement = min(max(new_arrangement, 0x45), 0x7f)
        self.map_data_arrangement[pos[1]][pos[0]] = new_arrangement
    
    
    def scroll_edit(self, mouse_event):
        for y in range(2):
            for x in range(2):
                mouse_x = int(mouse_event.scenePos().x() // 16) - x*16
                mouse_y = int(mouse_event.scenePos().y() // 16) - y*16
                pos = (self.world_x + x, self.world_y + y)
                if mouse_x >= 16-2 and mouse_x < 16+1 and mouse_y >= 0-1 and mouse_y < 16+1:
                    # right edge
                    self.scroll_toggle_flag(pos, 0x1)
                if mouse_x >= 0-1 and mouse_x < 0+2 and mouse_y >= 0-1 and mouse_y < 16+1:
                    # left edge
                    self.scroll_toggle_flag(pos, 0x2)
                if mouse_x >= 0-1 and mouse_x < 16+1 and mouse_y >= 0-1 and mouse_y < 0+2:
                    # up edge
                    self.scroll_toggle_flag(pos, 0x4)
                if mouse_x >= 0-1 and mouse_x < 16+1 and mouse_y >= 16-2 and mouse_y < 16+1:
                    # down edge
                    self.scroll_toggle_flag(pos, 0x8)
        self.world_update_interface()
                
    
    def scroll_toggle_flag(self, pos, flag):
        self.map_data_scroll[pos[1]][pos[0]] = self.map_data_scroll[pos[1]][pos[0]] ^ flag
    
    
    def map_file_save(self):
        file_content = "; Bank " + self.map_data_bank_number + " Level Data\n"
        file_content += ";  Saved using 2DEd\n"
        file_content += "SECTION \"ROM Bank $" + ("00" + self.map_data_bank_number)[-3:] + "\", ROMX[$4000], BANK[$" + self.map_data_bank_number + "]"
        file_content += "\n; Screen Data Pointers\n\n"
        for y in range(16):
            file_content += "    dw " + ", ".join(["${:02X}00".format(b) for b in self.map_data_arrangement[y]]) + "\n"
        file_content += "\n; Scroll Data\n\n"
        for y in range(16):
            file_content += "    db " + ", ".join(["${:02X}".format(b) for b in self.map_data_scroll[y]]) + "\n"
        file_content += "\n; Room Transition Indexes\n\n"
        for y in range(16):
            file_content += "    dw " + ", ".join(["${:03X}".format(w) for w in self.map_data_transition[y]]) + "\n"
        for s in range(0x45, 0x80):
            file_content += "\n; Screen ${:02X}00\n\n".format(s)
            for y in range(16):
                file_content += "    db " + ", ".join(["${:02X}".format(b) for b in self.map_data_screen[s][y]]) + "\n"

        with open(self.map_file_name, "w") as f:
            f.write(file_content)


    def enemy_file_load(self):
        try:
            names_file = []
            with open(self.src_folder + "/data/enemy_nameConstants.asm", "r") as file:
                for line in file.readlines():
                    try:
                        line = line[:line.index(";")]
                    except:
                        pass
                    names_file.append(line.strip())
            
            self.enemy_names = []
            self.enemy_ids = []
            for line in names_file:
                if line == "" or line == "\n":
                    continue
                line = [term.strip() for term in line.split("=")]
                self.enemy_names.append(line[0].split()[-1])
                self.enemy_ids.append(int(line[1].split()[0][1:], 16))
            
            enemy_file = []
            with open(self.src_folder + "/maps/enemyData.asm", "r") as file:
                for line in file.readlines():
                    if line == "\n" or line.startswith(";"):
                        continue
                    try:
                        line = line[:line.index(";")]
                    except:
                        pass
                    enemy_file.append(line.strip())
            enemy_file_void_index = next(i for i,line in enumerate(enemy_file) if not line.strip().startswith("dw "))
            enemy_file_banks = enemy_file[:enemy_file_void_index]
            enemy_file_dict = enemy_file[enemy_file_void_index:]
            
            enemy_dict = {}
            enemy_dict_keys = []
            enemy_dict_value = []
            for line in enemy_file_dict:
                enemy_data = ""
                line_has_db = True
                try:
                    _ = line.index("db ")
                except:
                    line_has_db = False
                    
                if line_has_db:
                    if line.startswith("db "):
                        enemy_data = line[line.index("db ")+len("db "):].strip()
                    else:
                        line = line.split(":")
                        enemy_dict_keys.append(line[0].strip())
                        enemy_data = line[1][line[1].index("db ")+len("db "):].strip()
                else:
                    line = line.split(":")
                    enemy_dict_keys.append(line[0].strip())
                
                if enemy_data != "":
                    enemy_data = [b.strip() for b in enemy_data.split(",")]
                    while len(enemy_data) >= 4:
                        if enemy_data[1].startswith("$"):
                            if int(enemy_data[1][1:], 16) not in self.enemy_ids:
                                raise Exception("The enemy id '" + enemy_data[1] + "' does not have a name in enemy_nameConstants.asm")
                            enemy_data[1] = self.enemy_names[self.enemy_ids.index(int(enemy_data[1][1:], 16))]
                        if enemy_data[1] not in self.enemy_names:
                            raise Exception("The enemy name '" + enemy_data[1] + "' is not defined in enemy_nameConstants.asm")
                        enemy = {
                            "respawn": int(enemy_data[0][1:], 16),
                            "name": enemy_data[1],
                            "x": int(enemy_data[2][1:], 16),
                            "y": int(enemy_data[3][1:], 16),
                        }
                        enemy_dict_value.append(enemy)
                        enemy_data = enemy_data[4:]
                    if len(enemy_data) == 1:
                        for key in enemy_dict_keys:
                            enemy_dict[key] = enemy_dict_value
                        enemy_dict_keys = []
                        enemy_dict_value = []
            
            enemy_banks_file_path = self.src_folder + "/maps/"
            enemy_banks_file_names = [f for f in os.listdir(enemy_banks_file_path) if os.path.isfile(os.path.join(enemy_banks_file_path, f))]
            self.enemy_banks = {}
            for f in enemy_banks_file_names:
                if not f.startswith("bank_"):
                    continue
                print(f)
                bank_number = "{:X}".format(int(os.path.splitext(f)[0].split("bank_")[1], 16))
                self.enemy_banks[bank_number] = []
            
            
            for _, bank in self.enemy_banks.items():
                for y in range(16):
                    if len(enemy_file_banks) == 0:
                        raise Exception("The enemy screens do not have enough rows in enemyData.asm")
                    line = enemy_file_banks.pop(0)
                    line = [b.strip() for b in line[len("dw "):].split(",")]
                    if len(line) != 16:
                        raise Exception("The enemy screen row is of length " + len(line) + " instead of 16 in enemyData.asm")
                    for x in range(len(line)):
                        if line[x] == "enVoid":
                            line[x] = None
                        elif line[x] in enemy_dict:
                            line[x] = enemy_dict[line[x]]
                        else:
                            raise Exception("The enemy screen name '" + line[x] + "' is not defined in enemyData.asm")
                    bank.append(line)
            if len(enemy_file_banks) != 0:
                raise Exception("The enemy screens have too many rows in enemyData.asm")
            
            self.update_interface()
        except Exception as ex:
            messageBox = QtWidgets.QMessageBox(self)
            messageBox.setWindowTitle("Failed to load enemy file")
            messageBox.setText(str(ex))
            messageBox.setDetailedText(str(traceback.format_exc()))
            messageBox.setIcon(QtWidgets.QMessageBox.Icon.Critical)
            messageBox.exec()
            sys.exit()
    
    
    def enemy_edit(self, mouse_event):
        if self.enemy_banks == None or self.map_data_bank_number == None:
            return
            
        world_x = self.world_x + int(mouse_event.scenePos().x() // 256)
        world_y = self.world_y + int(mouse_event.scenePos().y() // 256)
        
        enemy_screen = self.enemy_banks[self.map_data_bank_number][world_y][world_x]
        
        if enemy_screen == None:
            return
            
        permanent_enemies = []
        for row in self.enemy_banks[self.map_data_bank_number]:
            for screen in row:
                if screen == None:
                    continue
                for enemy in screen:
                    if enemy["respawn"] >= 0x40:
                        permanent_enemies.append(enemy)
        
        permanent_enemies.sort(key=lambda enemy: enemy["respawn"])
        for enemy in permanent_enemies:
            print("${:02X}: ".format(enemy["respawn"]) + enemy["name"])
        
        dialog = Interface_EnemyEdit([enemy.copy() for enemy in enemy_screen], self.enemy_names)
        dialog.setWindowTitle("Edit enemy data (Bank " + self.map_data_bank_number + ", Screen x" + str(world_x) + " y" + str(world_y) + ")")
        dialog.setFixedWidth(500)
        dialog.setMinimumHeight(500)
        if not dialog.exec():
            return
            
        self.enemy_banks[self.map_data_bank_number][world_y][world_x] = dialog.enemy_screen
        self.update_interface()
    
    
    def enemy_file_save(self):
        file_content = "; in bank 3\n\n"
        
        enemy_dict = {}
        
        for bank_number, bank in self.enemy_banks.items():
            file_content += "; Enemy Data Pointers for Bank " + bank_number + "\n"
            for y, row in enumerate(bank):
                row = row.copy()
                for x, item in enumerate(row):
                    if item == None:
                        row[x] = "enVoid"
                    elif len(item) < 1:
                        row[x] = "enNone"
                    else:
                        key = "en" + bank_number + "_" + "{:X}".format(y) + "{:X}".format(x)
                        row[x] = key
                        enemy_dict[key] = item
                file_content += "    dw " + ", ".join(row) + "\n"
            file_content += "\n"
        
        file_content += """; Enemy Data
enVoid: 
enNone: 
    db ENEMY_DATA_END

; my notes about this format: enemy respawn id (40-7F is saved in sram), enemy sprite id, x, y
"""
        for key, enemy_list in enemy_dict.items():
            file_content += key + ": \n"
            for enemy in enemy_list:
                file_content += "    db " 
                file_content += "${:02X}, ".format(enemy["respawn"])
                file_content += enemy["name"] + ", "
                file_content += "${:02X}, ".format(enemy["x"])
                file_content += "${:02X}\n".format(enemy["y"])
            file_content += "    db ENEMY_DATA_END\n"
        
        with open(self.src_folder + "/maps/enemyData.asm", "w") as f:
            f.write(file_content)

